from ..imports import *
from .config_utils import *
from .db_functions import *
from .get_tables import *
from .get_templates import *
from .get_time_data import *
from .image_utils import *
from .legacy_utils import *
from .solana import *
def select_one(query, *args):
    rows = select_rows(query, *args)
    return get_rows(rows)



from abstract_solcatcher_database import *


def fetchmeta(mint):
    meta_data = asyncio.run(async_call_solcatcher_ts('get-metadata-foundation',mint=mint,url=ankr_url,get_id=False))
    
    return get_meta_data_from_meta_id(meta_data)
def get_or_fetch_meta(mint):
    meta_data = asyncio.run(async_call_solcatcher_ts('get-or-fetch-metadata',mint=mint))
    return meta_data



from .baseQueryManager import *
from .columnNamesManager import *
from .connectionManager import *
from .databaseBrowser import *
from .databaseManager import *
from .dbManager import *
from .envManager import *
from .sessionManager import *
from .tableManager import *



from .databaseManager import *
def ensure_db_manager(db_mgr=None, env_path=None, dbName=None, dbType=None, conn_mgr=None):
    return db_mgr or DatabaseManager(env_path=env_path, dbName=dbName, dbType=dbType, conn_mgr=conn_mgr)



from .utils import *
class connectionManager(metaclass=SingletonMeta):
        
    def __init__(self, env_path=None, dbType=None, dbName=None, tables=[], tables_path=None,dbVars=None):
        if not hasattr(self, 'initialized'):
            self.initialized=True
            self.env_path = self.get_env_path(env_path=env_path)
            
            self.dbName = get_dbName(dbName=dbName)
            
            self.dbType = get_dbType(dbType=dbType)
            
            self.dbVars = dbVars or self.get_db_vars(dbType=self.dbType, dbName=self.dbName, env_path=self.env_path)
            self.user = self.dbVars['user']
            self.password = self.dbVars['password']
            self.host = self.dbVars['host']
            self.port = self.dbVars['port']
            self.dbname = self.dbVars['dbname']
            self.dburl = self.dbVars['dburl']  # URL-based connection string
            self.table_mgr = TableManager()
            self.tables = tables or safe_load_from_json(file_path=tables_path) or []
            self.table_mgr.env_path = self.env_path
            self.add_insert_list=None
            
            self.check_conn()
        
    def check_conn(self):
        if self.add_insert_list == None:
          try:
                self.table_mgr.add_insert_list(self.connect_db(), self.tables, self.dbName)
                self.table_mgr.add_insert_list(self.connect_db(), self.tables, self.dbName)
                self.add_insert_list=True
          except:
            pass
        return self.add_insert_list
    def get_dbName(self, dbName=None):
        return get_dbName(dbName=dbName or self.dbName)
    def get_dbType(self, dbType=None):
        return get_dbType(dbType=dbType or self.dbType)
    def get_env_path(self, env_path=None):
        return verify_env_path(env_path=env_path)

    def get_db_vars(self, env_path=None, dbType=None, dbName=None):
        env_path = self.get_env_path(env_path=env_path)
        dbName = self.get_dbName(dbName=dbName)
        dbType = self.get_dbType(dbType=dbType)
        dbVars = get_db_vars(env_path=env_path, dbType=dbType, dbName=dbName)
        return dbVars

    def change_db_vars(self, env_path=None, dbType=None, dbName=None, tables=[]):
        self.env_path = self.get_env_path(env_path=env_path)
        self.dbName = self.get_dbName(dbName=dbName)
        
        self.dbType = self.get_dbType(dbType=dbType)
        self.dbVars = self.get_db_vars(env_path=self.env_path, dbType=dbType, dbName=dbName)
        self.user = self.dbVars['user']
        self.password = self.dbVars['password']
        self.host = self.dbVars['host']
        self.port = self.dbVars['port']
        self.dbname = self.dbVars['dbname']
        self.dburl = self.dbVars['dburl']
        self.simple_connect = self.simple_connect_db()
        self.get_db_connection(self.connect_db())
        self.tables = tables or self.tables
        self.table_mgr.add_insert_list(self.connect_db(), self.tables, self.dbName)
        return self.dbVars

    def connect_db(self):
            
            """ Establish a connection to the database, either by connection parameters or via URL """
            if self.dburl:
                
                return psycopg2.connect(self.dburl)
            else:
                return psycopg2.connect(user=self.user,
                                        password=self.password,
                                        host=self.host,
                                        port=self.port,
                                        dbname=self.dbname)

    def simple_connect_db(self):
        """ Create a connection pool using the database URL """
        if self.dburl:
            return psycopg2.pool.SimpleConnectionPool(1, 10, self.dburl)
        else:
            return psycopg2.pool.SimpleConnectionPool(1, 10, user=self.user,
                                                      password=self.password,
                                                      host=self.host,
                                                      port=self.port,
                                                      database=self.dbname)

    def put_db_connection(self, conn):
        conn = conn or self.connect_db()
        self.putconn(conn)

    def get_db_connection(self):
        return self.connect_db()

    def get_insert(self, tableName):
        return self.table_mgr.get_insert(tableName)

    def fetchFromDb(self, tableName, searchValue):
        return self.table_mgr.fetchFromDb(tableName, searchValue, self.connect_db())

    def insertIntoDb(self, tableName, searchValue, insertValue):
        return self.table_mgr.insert_intoDb(tableName, searchValue, insertValue, self.connect_db())

    def search_multiple_fields(self, query, **kwargs):
        return self.table_mgr.search_multiple_fields(query=query, conn=self.connect_db())

    def get_first_row_as_dict(self, tableName=None, rowNum=1):
        return self.table_mgr.get_first_row_as_dict(tableName=tableName, rowNum=rowNum, conn=self.connect_db())



from abstract_utilities import SingletonMeta
class sessionManager(metaclass=SingletonMeta):
    def __init__(self, Base=None,db_vars=None,dbDir=None):
        self.dbTracker = {}
        self.Base = Base or declarative_base()
        self.db_vars=db_vars or {}
        self.dbDir=dbDir or os.path.join(os.getcwd(),"databases")
        self.initialize_all_dbs()
    def initialize_all_dbs(self):
        for dbName, db_info in self.db_vars.items():
            dbPath = os.path.join(self.dbDir,f"{dbName}.db")
            dbName = get_db_name(dbPath=dbPath, dbUrl=get_db_url(dbPath))
            self.initialize_db(dbName=dbName,dbPath=dbPath, db_tables=db_info)
    def initialize_db(self,dbName=None, dbPath=None, dbUrl=None, db_tables=None):
        dbUrl = dbUrl or get_db_url(dbPath)
        
        self.checkDatabaseName(dbName=dbName, dbPath=dbPath, dbUrl=dbUrl, db_tables=self.db_vars[dbName])
        self.create_session(dbPath=dbPath, dbUrl=dbUrl)
        self.create_tables(dbName)
        return dbName

    def get_dbName(self, dbName=None, dbPath=None, dbUrl=None):
        if dbName is None and dbUrl is None and dbPath is None:
            return dbName
        elif dbName is None and (dbUrl is not None or dbPath is not None):
            dbName = get_db_name(dbPath=dbPath, dbUrl=dbUrl)
        return dbName

    def create_session(self, dbPath=None, dbUrl=None):
        dbName = get_db_name(dbPath=dbPath, dbUrl=dbUrl)
        self.dbTracker[dbName]["engine"] = get_db_engine(dbUrl=dbUrl, dbPath=dbPath)
        self.Base.metadata.bind = self.dbTracker[dbName]["engine"]
        self.solcatcherDBSession = sessionmaker(bind=self.dbTracker[dbName]["engine"])
        self.dbTracker[dbName]["session"] = self.solcatcherDBSession()

    def create_tables(self, dbName=None, dbPath=None, dbUrl=None):
        self.Base.metadata.create_all(self.dbTracker[dbName]["engine"])
        dbName = get_db_name(dbName=dbName, dbPath=dbPath, dbUrl=dbUrl)
        for table, columns_info in self.dbTracker[dbName]["db_tables"].items():
            columns = list(columns_info["valueKeys"].keys())
            columns_defs = self.flatten_columns_defs(columns_info["valueKeys"])
            create_table_sql = text(f'CREATE TABLE IF NOT EXISTS "{table}" ({columns_defs});')
            self.dbTracker[dbName]["session"].execute(create_table_sql)
        self.dbTracker[dbName]["session"].commit()


    def flatten_columns_defs(self, valueKeys):
        column_defs = []
        for col, dtype in valueKeys.items():
            col_name = f'"{col}"'
            if isinstance(dtype, dict):
                column_defs.append(f"{col_name} JSON")
            elif isinstance(dtype, list):
                column_defs.append(f"{col_name} JSON")
            else:
                column_defs.append(f"{col_name} {dtype}")
        return ", ".join(column_defs)

    def checkDatabaseName(self, dbName=None, dbPath=None, dbUrl=None, dbBrowser=None, db_tables=None, primary_types=None):
        dbName = self.get_dbName(dbName=dbName, dbPath=dbPath, dbUrl=dbUrl)
        if dbName not in self.dbTracker:
            self.dbTracker[dbName] = {"dbUrl": dbUrl, "dbPath": dbPath, "db_tables": db_tables, "columns_info": self.db_vars.get(dbName),"classes":{}}
            for key,value in self.db_vars.get(dbName).items():
                self.dbTracker[dbName]["classes"][key] = create_class_from_dict(capitalize(key), value,self.Base)
                
    def close_session(self, dbName=None, dbPath=None, dbUrl=None):
        dbName = self.get_dbName(dbName=dbName, dbPath=dbPath, dbUrl=dbUrl)
        self.dbTracker[dbName]["session"].close()

    # Function to insert a record
    def insert_record(self, dbName, class_name, record, unique_keys=None):
        model = self.dbTracker[dbName]["classes"][class_name]
        session = self.dbTracker[dbName]["session"]
        unique_keys=make_list(unique_keys or self.dbTracker[dbName]["columns_info"][class_name]['unique_keys'])
        if record_exists(session, model, **{key: record[key] for key in unique_keys}):
            #print(f"Duplicate entry found with keys: {unique_keys}. Skipping...")
            return
        new_record = model(**record)
        try:
            session.add(new_record)
            session.commit()
        except IntegrityError:
            session.rollback()
            print(f"Integrity error for record: {record}. Rolling back transaction.")


    # Function to update a record
    def update_record(self, dbName, table_name, record, conditions):
        session = self.dbTracker[dbName]["session"]
        set_clause = ", ".join([f'"{key}" = :{key}' for key in record.keys()])
        condition_clause = " AND ".join([f'"{key}" = :{key}' for key in conditions.keys()])
        update_sql = text(f'UPDATE "{table_name}" SET {set_clause} WHERE {condition_clause}')
        session.execute(update_sql, {**record, **conditions})
        session.commit()

    # Function to delete a record
    def delete_record(self, dbName, table_name, conditions):
        session = self.dbTracker[dbName]["session"]
        condition_clause = " AND ".join([f'"{key}" = :{key}' for key in conditions.keys()])
        delete_sql = text(f'DELETE FROM "{table_name}" WHERE {condition_clause}')
        session.execute(delete_sql, conditions)
        session.commit()



from ..queries import query_data_as_dict, query_data, derive_timestamp

def fetch_filtered_transactions_paginated_old(
    sol_amount, 
    operator=">",
    timestamp_operator=None, 
    limit=None, 
    offset=0,
    years=0,
    months=0,
    weeks=0,
    days=0,
    hours=0,
    minutes=0,
    seconds=0,
    timestamp=None,
    *args,
    **kwargs
):
    # Build the base query and include a LEFT JOIN to get tcn counts per pair.
    base_query = f"""
        FROM 
            transactions t
        JOIN 
            pairs p ON t.pair_id = p.id
        LEFT JOIN 
            metadata m ON p.meta_id = m.id
        LEFT JOIN (
            SELECT 
                pair_id, 
                SUM(jsonb_array_length(tcns)) AS tcn_count
            FROM 
                transactions
            GROUP BY 
                pair_id
        ) tcns_count ON p.id = tcns_count.pair_id
        WHERE 
            t.signature IN (
                SELECT signature FROM pairs WHERE signature IS NOT NULL
            )
        AND 
            t.program_id = p.program_id
        AND 
            EXISTS (
                SELECT 1
                FROM jsonb_array_elements(t.tcns) AS elem
                WHERE (elem ->> 'sol_amount')::numeric {operator} %s
    """
    params = [sol_amount]

    ts = derive_timestamp(
        years=years,
        months=months,
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        timestamp=timestamp
    )
    
    if ts and timestamp_operator:
        base_query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(ts))
    
    # Close the EXISTS clause.
    base_query += ")"
    if metaDataKeys:
        # Optionally, validate the keys against an allowed list to prevent SQL injection.
        for key, flag in metaDataKeys.items():
            if flag:
                # True: value should not be None or false
                query += f" AND (m.{key} IS NOT NULL AND m.{key} <> false)"
            else:
                # False: value should be either None or false
                query += f" AND (m.{key} IS NULL OR m.{key} = false)"
    
    query += " ORDER BY t.updated_at DESC LIMIT %s OFFSET %s;"
    # Build the final data query including the tcn_count column.
    data_query = f"""
        SELECT
            p.*, 
            m.*,
            t.*,
            COALESCE(tcns_count.tcn_count, 0) AS tcn_count
        {base_query}
        ORDER BY t.updated_at DESC
        LIMIT %s OFFSET %s;
    """
    data_params = params + [limit, offset]
    transactions = query_data_as_dict(data_query, data_params)
    
    return transactions
def fetch_filtered_transactions_paginated(
        sol_amount, 
        operator=">",
        timestamp_operator=None, 
        limit=None, 
        offset=0,
        years=0,
        months=0,
        weeks=0,
        days=0,
        hours=0,
        minutes=0,
        seconds=0,
        timestamp=None,
        metaDataKeys=None,
        *args,
        **kwargs
    ):
        # Build the base query and include a LEFT JOIN to get tcn counts per pair.
        base_query = f"""
            FROM 
                transactions t
            JOIN 
                pairs p ON t.pair_id = p.id
            LEFT JOIN 
                metadata m ON p.meta_id = m.id
            LEFT JOIN (
                SELECT 
                    pair_id, 
                    SUM(jsonb_array_length(tcns)) AS tcn_count
                FROM 
                    transactions
                GROUP BY 
                    pair_id
            ) tcns_count ON p.id = tcns_count.pair_id
            WHERE 
                t.signature IN (
                    SELECT signature FROM pairs WHERE signature IS NOT NULL
                )
            AND 
                t.program_id = p.program_id
            AND 
                EXISTS (
                    SELECT 1
                    FROM jsonb_array_elements(t.tcns) AS elem
                    WHERE (elem ->> 'sol_amount')::numeric {operator} %s
        """
        params = [sol_amount]

        ts = derive_timestamp(
            years=years,
            months=months,
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            timestamp=timestamp
        )
        
def fetch_filtered_transactions_paginated(
    sol_amount, 
    operator=">",
    timestamp_operator=None, 
    limit=None, 
    offset=0,
    years=0,
    months=0,
    weeks=0,
    days=0,
    hours=0,
    minutes=0,
    seconds=0,
    timestamp=None,
    metaDataKeys=None,
    *args,
    **kwargs
):
    # Build the base query and include a LEFT JOIN to get tcn counts per pair.
    base_query = f"""
        FROM 
            transactions t
        JOIN 
            pairs p ON t.pair_id = p.id
        LEFT JOIN 
            metadata m ON p.meta_id = m.id
        LEFT JOIN (
            SELECT 
                pair_id, 
                SUM(jsonb_array_length(tcns)) AS tcn_count
            FROM 
                transactions
            GROUP BY 
                pair_id
        ) tcns_count ON p.id = tcns_count.pair_id
        WHERE 
            t.signature IN (
                SELECT signature FROM pairs WHERE signature IS NOT NULL
            )
        AND 
            t.program_id = p.program_id
        AND 
            EXISTS (
                SELECT 1
                FROM jsonb_array_elements(t.tcns) AS elem
                WHERE (elem ->> 'sol_amount')::numeric {operator} %s
    """
    params = [sol_amount]

    ts = derive_timestamp(
        years=years,
        months=months,
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        timestamp=timestamp
    )
    
    if ts:
        if timestamp_operator == None:
            timestamp_operator = ">"
        base_query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(ts))
    
    # Close the EXISTS clause.
    base_query += ")"
    
    # Add metadata filtering conditions.
    if metaDataKeys:
        for key, filter_val in metaDataKeys.items():
            # Skip any global options.
            if key == '__option':
                continue

            # Use a CASE expression that only casts to JSON if the text looks like valid JSON.
            # Here we check that the first two characters equal '{"'
            json_case = (
                f"(CASE WHEN m.{key} IS NOT NULL AND substring(m.{key} from 1 for 2) = '{{\"' "
                f"THEN m.{key}::jsonb ->> '__option' ELSE m.{key} END)"
            )
            
            if isinstance(filter_val, dict):
                option_value = filter_val.get('__option')
                if option_value == "None" or option_value is None:
                    # Require that the value is NULL or (if JSON) its '__option' equals 'None'.
                    base_query += f" AND (m.{key} IS NULL OR {json_case} = 'None')"
                else:
                    base_query += f" AND (m.{key} IS NOT NULL AND {json_case} <> %s)"
                    params.append(option_value)
            else:
                # When filter_val is a boolean.
                if filter_val:
                    base_query += f" AND (m.{key} IS NOT NULL AND {json_case} NOT IN ('None','false'))"
                else:
                    base_query += f" AND (m.{key} IS NULL OR {json_case} IN ('None','false'))"
    
    # Build the final data query including the tcn_count column.
    data_query = f"""
        SELECT
            p.*, 
            m.*,
            t.*,
            COALESCE(tcns_count.tcn_count, 0) AS tcn_count
        {base_query}
        ORDER BY t.updated_at DESC
        LIMIT %s OFFSET %s;
    """
    data_params = params + [limit, offset]
    transactions = query_data_as_dict(data_query, data_params)
    
    return transactions




from abstract_utilities import make_list,SingletonMeta,get_logFile
from ....query_utils.utils.query_utils.utils import query_data,get_all_table_names
from psycopg2 import sql, connect
logger = get_logFile('query_utils')



from .main import *
def get_column_names(tableName,schema='public'):
    return columnNamesManager().get_column_names(tableName,schema)
def getZipRows(tableName, rows, schema='public'):
    columnNames = get_column_names(tableName,schema)
    if columnNames:
        return [dict(zip(columnNames,row)) for row in make_list(rows) if row]



from ..imports import *
logger = get_logFile('query_utils')
def get_connection():
    return connect_db()
def execute_query(query, values=None, fetch=True, as_dict=True):
    """
    Execute a SQL query and return results if applicable.
    
    Args:
        query (str or psycopg2.sql.Composed): SQL query to execute.
        values (tuple, optional): Values for parameterized queries.
        fetch (bool): Whether to fetch results (for SELECT) or commit (for INSERT/UPDATE).
        as_dict (bool): Return results as dictionaries if True, else as tuples.
    
    Returns:
        list: Query results (empty if no fetch or error).
    """
    # Convert Composed query to string if necessary
    if isinstance(query, sql.Composed):
        query_str = query.as_string(get_connection())
    else:
        query_str = str(query)

    logger.info(f"Executing query: {query_str} with values: {values}")
    conn = get_connection()
    cursor_factory = RealDictCursor if as_dict else None
    
    try:
        with conn.cursor(cursor_factory=cursor_factory) as cursor:
            cursor.execute(query, values)
            if fetch and query_str.strip().upper().startswith("SELECT"):
                result = cursor.fetchall()
                if result:
                    logger.debug(f"First row: {result[0]}")
                return result
            conn.commit()
            return []
    except Exception as e:
        conn.rollback()
        logger.error(f"Query failed: {query_str}\nValues: {values}\nError: {e}\n{traceback.format_exc()}")
        return []
    finally:
        conn.close()

def get_all_table_names(schema='public'):
    """Fetch all table names from a specified schema."""
    query = """
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = %s AND table_type = 'BASE TABLE';
    """
    result = execute_query(query, values=(schema,), fetch=True, as_dict=False)
    return [row[0] for row in result] if result else []

# Legacy functions with deprecation warnings
def query_data_as_dict(query, values=None, error="Error executing query:"):
    warnings.warn("query_data_as_dict is deprecated; use execute_query instead.", DeprecationWarning)
    return execute_query(query=query, values=values, fetch=True, as_dict=True)

def get_query_result(query, values=None, zipit=False, **kwargs):
    warnings.warn("get_query_result is deprecated; use execute_query instead.", DeprecationWarning)
    return execute_query(query, values=values, fetch=True, as_dict=zipit)

def query_data(query, values=None, error="Error executing query:", zipRows=True):
    warnings.warn("query_data is deprecated; use execute_query instead.", DeprecationWarning)
    logger.info(f"query = {query} and values = {values}")
    return execute_query(query, values=values, fetch=True, as_dict=zipRows)

def aggregate_rows(query, values=None, errorMsg='Error Fetching Rows', fetch=True, as_dict=None, zipRows=None, zipit=None, **kwargs):
    warnings.warn("aggregate_rows is deprecated; use execute_query instead.", DeprecationWarning)
    # Resolve as_dict from multiple possible parameters
    resolved_as_dict = as_dict if as_dict is not None else (zipRows if zipRows is not None else (zipit if zipit is not None else True))
    return execute_query(query, values=values, fetch=fetch, as_dict=resolved_as_dict)



from ..imports import *
def select_one(query, *args):
    rows = select_rows(query, *args)
    return get_rows(rows)
class BaseQueryManager(metaclass=SingletonMeta):
    """
    Generic query manager: load <basename>.json for named queries,
    provide generic insert/update helpers, plus run() for custom SQL.
    """
    def __init__(self, basename: str, logs_on: bool = True):
        if getattr(self, "_initialized", False):
            return
        self._initialized = True

        self.logs_on = logs_on
        self.toggle_trigger = {True: False, False: True}

        # load your JSON / YAML for this manager
        data = get_yaml_queries_data(basename)
        self._queries = data

        # dynamically set self._query_<key> = the SQL string
        for key, sql in data.items():
            setattr(self, f"_query_{key}", sql)
    def display_query_options(self) -> None:
        for key,value in self._queries.items():
            print(f"{key}:\n{value}\n\n")
    def toggle_logs(self, toggle: Optional[bool] = None) -> None:
        self.logs_on = not self.logs_on if toggle is None else bool(toggle)

    def insert(self, table: str, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Generic INSERT: builds SQL from data dict, returns inserted row.
        """
        cols = list(data.keys())
        vals = [data[c] for c in cols]
        placeholders = ", ".join(["%s"] * len(cols))
        col_list = ", ".join(cols)
        sql = f"INSERT INTO {table} ({col_list}) VALUES ({placeholders}) RETURNING *;"
        return select_one(sql, *vals)

    def update(self, table: str, data: Dict[str, Any], where: Dict[str, Any]) -> Dict[str, Any]:
        """
        Generic UPDATE: set columns from data dict, filter by where dict, returns updated row.
        """
        set_cols = list(data.keys())
        set_expr = ", ".join([f"{col} = %s" for col in set_cols])
        set_vals = [data[c] for c in set_cols]

        where_cols = list(where.keys())
        where_expr = " AND ".join([f"{col} = %s" for col in where_cols])
        where_vals = [where[c] for c in where_cols]

        sql = f"UPDATE {table} SET {set_expr} WHERE {where_expr} RETURNING *;"
        return select_one(sql, *(set_vals + where_vals))

    def run(
        self,
        key: str,
        *args,
        one: bool = False,
        many: bool = False,
        commit: bool = False,
        returning: bool = False,
        map_fn: Optional[Any] = None,
    ) -> Any:
        """
        Execute a query by its key from the loaded YAML/JSON.
        """
        sql = self._queries.get(key)
        if not sql:
            raise KeyError(f"Query '{key}' not found in configuration.")

        if self.logs_on:
            initialize_call_log()

        if many:
            rows = select_distinct_rows(sql, *args)
            return map_fn(rows) if map_fn else rows

        if one:
            row = select_rows(sql, *args)
            return map_fn(row) if (map_fn and row) else row

        if returning:
            return insert_query(sql, *args)

        if commit:
            execute_query(sql, *args)
            return None

        return select_distinct_rows(sql, *args)



from ..imports import *
def make_single(string):
  return string.replace('_','')
def make_multiple(string):
    nustring=''
    
    for char in string:
        if char in uppers:
            char = f"_{char.lower()}"
        nustring+=char
    return nustring
def get_table(method,tables):
  tableName = make_single(method.lower())
  table = [table for table in tables if table.get('tableName').lower() == tableName]
  if table and isinstance(table,list):
    table = table[0]
  return table
def get_db_vars(dbName, dbType=None,env_path=None):
    dbType='database'
    dbVars = {"user": None, "dbname": None, "host": None, "port": None, "password": None}
    for key in dbVars:
        env_key = f"{dbName.upper()}_{dbType.upper()}_{key.upper()}"
        dbVars[key] = get_env_value(key=env_key,path=get_env_path())
    return dbVars
def get_db_url(dbName, dbType=None):
    dbVars = get_db_vars(dbName, dbType=dbType)
    return f"postgresql://{dbVars['user']}:{dbVars['password']}@{dbVars['host']}:{dbVars['port']}/{dbVars['dbname']}"
def get_engine(dbName):
    db_url = get_db_url(dbName)
    engine = create_engine(db_url)
    return engine



import logging
logging.basicConfig()
from abstract_math import divide_it
from abstract_utilities import get_any_value
def get_if_list(obj):
    if obj and isinstance(obj,list):
        return obj[0]
def get_mint_length():
    mint_length = len('E3HurESTBHNed6HnARSMBg45i9va2d9BYeZMcnsQpump')
    return mint_length
def get_signature_length():
    signature = len('34WyijxfHuoabvHpmHFNxrBvGwRXy12vdtbDvt4bQThdALEWyRYqN39VMyncXC3FoYK8BVBP39ahws7z615jZ45T')
    return signature
def is_mint(obj):
    if isinstance(obj,str):
        return len(obj) == get_mint_length()
    return False
def is_signature(obj):
    if isinstance(obj,str):
        return len(obj) == get_signature_length()
    return False
def if_list_get_single(obj):
    while True:
        if obj != None and isinstance(obj,list) and len(obj) == 1:
            obj = obj[0]
        else:
            break
    
    return obj
def get_zero_or_real(obj):
    while True:
        if isinstance(obj,list) and len(obj) == 1:
            obj = if_list_get_single(obj)
        else:
            break
    if isinstance(obj,list) and len(obj) > 1:
        obj = obj[0]
    return obj or 0
def get_key_value(txn,key):
    if isinstance(txn, list):
        if txn and 'tcns' in txn[0]:
            return get_any_value(txn[0]['tcns'], key)
        else:
            txn = if_list_get_single(txn)
    return get_any_value(txn, key)
def flatten_txn(txn):
    txns = []
    for tcn in get_any_value(txn,'tcns') or []:
        new_txn = {**txn, **tcn}
        if 'tcns' in new_txn:
            del new_txn['tcns']
        txns.append(new_txn)
    return if_list_get_single(txns or txn)
def flatten_txn_history(txns):
    return [flatten_txn(txn) for txn in txns or [] if txn]
def sort_txn_history(txns):
    sorted_txns = sorted(txns, key=lambda txn: get_timestamps(txn))
    return sorted_txns
def get_sorted_txn_history(txns):
    txns = flatten_txn_history(txns)
    sorted_txns = sort_txn_history(txns)
    return sorted_txns
def get_string_value(txn,key):
    value = get_key_value(txn,key)
    if isinstance(value,list):
        value = list(set(value))
    return if_list_get_single(value)
def get_integer_value(txn,key):
    value = get_key_value(txn,key)
    return get_zero_or_real(value)
def get_bool_value(txn,key):
    value = get_key_value(txn,key)
    return if_list_get_single(value)
def get_txn_price(txn):
    return get_integer_value(txn,'price')
def get_virtual_sol_reserves(txn):
    return get_key_value(txn,'virtual_sol_reserves')
def get_virtual_token_reserves(txn):
    return get_key_value(txn,'virtual_token_reserves')
def get_user_address(txn):
    return get_string_value(txn,'user_address')
def get_mint(txn):
    return get_string_value(txn,'mint')
def get_bonding_curve(txn):
    return get_string_value(txn,'bonding_curve')
def get_timestamps(txn):
    return get_integer_value(txn,'timestamp')
def get_sol_amount(txn):
    return get_integer_value(txn,'sol_amount')
def get_supply(txn):
    value = get_integer_value(txn,'supply')
    if not value:
        value = get_key_value(txn,'supply')
    return value
def get_token_amount(txn):
    return get_integer_value(txn,'token_amount')
def get_is_buy(txn):
    return get_bool_value(txn,'is_buy')
def get_signature(txn):
    return get_string_value(txn,'signature')
def get_all_timestamps(txns):
    timestamps = [get_timestamps(txn) for txn in txns]
    timestamps.sort()
    return timestamps
def get_price(virtualSolReserves, virtualTokenReserves):
    return divide_it(virtualSolReserves,virtualTokenReserves)
def get_any_function(txn,key):
    result=None
    key_lower = key.lower().replace(' ','_')
    funcs_js = {'price': get_txn_price,
     'virtual_sol_reserves': get_virtual_sol_reserves,
     'virtual_token_reserves': get_virtual_token_reserves,
     'user_address': get_user_address,
     'mint': get_mint,
     'bonding_curve': get_bonding_curve,
     'timestamp': get_timestamps,
     'sol_amount': get_sol_amount,
     'supply': get_supply,
     'token_amount': get_token_amount,
     'is_buy': get_is_buy,
     'signature':get_signature}
    func = funcs_js.get(key_lower)
    if func:
        result = func(txn)
    
    else:
        flattened_txn= flatten_txn(txn)
        flattened_txn = get_if_list(flattened_txn)
        if flattened_txn:
            result = get_key_value(flattened_txn,key)
            if result in [None,[]]:
                result = flattened_txn.get(key)
    return result

    



from ..imports import *
def get_table_name(tableName,schema='public'):
    return sql.SQL('{}.{}').format(
        sql.Identifier(schema),
        sql.Identifier(tableName)
        )
def _build_set_clause(data_map):
    """
    Returns: (sql.SQL fragment, values list)
    """
    parts, vals = [], []
    for col, val in data_map.items():
        parts.append(sql.SQL("{} = %s").format(sql.Identifier(col)))
        vals.append(val)
    return sql.SQL(', ').join(parts), vals


def _build_where_clause(filter_map, values=None, *, any_value=False):
    """
    Returns: (sql.SQL fragment starting with ' WHERE …' or sql.SQL(''), values list)
    """
    if not filter_map:
        return sql.SQL(''), values or []

    parts, vals = [], values or []
    for col, val in filter_map.items():
        if any_value:
            parts.append(sql.SQL("{} = ANY(%s)").format(sql.Identifier(col)))
            vals.append(make_list(val))
        else:
            parts.append(sql.SQL("{} = %s").format(sql.Identifier(col)))
            vals.append(val)

    return sql.SQL(' WHERE ') + sql.SQL(' AND ').join(parts), vals

# ---------- SELECT -----------------------------------------------------------
def fetch_any_combo(*,
                    column_names='*',
                    table_name,
                    search_map=None,
                    count=False,
                    any_value=False,
                    zipit=True,
                    schema='public'):

    if not table_name or table_name == '*':
        logger.error("Invalid table_name provided to fetch_any_combo: %s", table_name)
        return []

    # SELECT list
    if count:
        select_cols = sql.SQL("COUNT(*)")
    elif column_names == '*':
        select_cols = sql.SQL('*')
    else:
        cols = [c.strip() for c in column_names.split(',')]
        select_cols = sql.SQL(', ').join(sql.Identifier(c) for c in cols)

    base = sql.SQL("SELECT {} FROM {}.{}").format(
        select_cols,
        sql.Identifier(schema),
        sql.Identifier(table_name)
    )

    where_sql, params = _build_where_clause(search_map or {}, any_value=any_value)

    return query_data(base + where_sql, values=params, zipRows=zipit)


# ---------- UPDATE -----------------------------------------------------------
def update_any_combo(*,
                     table_name: str,
                     update_map: dict,
                     search_map: dict = None,
                     any_value: bool = False,
                     returning=False,        # False | True | 'count' | 'col1,col2'
                     zipit=True,
                     schema='public'):

    if not table_name or table_name == '*':
        raise ValueError("table_name is required")
    if not update_map:
        raise ValueError("update_map cannot be empty")

    set_sql, params = _build_set_clause(update_map)
    where_sql, params = _build_where_clause(search_map or {}, params, any_value=any_value)

    qry = (
        sql.SQL("UPDATE {}.{} SET ").format(sql.Identifier(schema),
                                            sql.Identifier(table_name))
        + set_sql
        + where_sql
    )

    if returning:
        if returning is True:
            qry += sql.SQL(' RETURNING *')
        elif returning == 'count':
            qry = sql.SQL("WITH upd AS (") + qry + sql.SQL(" RETURNING 1) SELECT COUNT(*) FROM upd")
        else:
            cols = [c.strip() for c in returning.split(',')]
            qry += sql.SQL(' RETURNING ') + sql.SQL(', ').join(sql.Identifier(c) for c in cols)

    return query_data(qry, values=params, zipRows=zipit)
def get_column_names(tableName,schema='public'):
    return columnNamesManager().get_column_names(tableName,schema)
def getZipRows(tableName, rows, schema='public'):
    columnNames = get_column_names(tableName,schema)
    if columnNames:
        return [dict(zip(columnNames,row)) for row in make_list(rows) if row]
def get_db_from(tableName=None,columnNames=None,searchColumn=None,searchValue=None,count=False,zipit=True):
    columnNames=columnNames or '*'
    if isinstance(columnNames,list):
        columnNames = ','.join(columnNames)
    response = fetch_any_combo(tableName=tableName,columnNames=columnNames,searchColumn=searchColumn,searchValue=searchValue,zipit=zipit,count=count)
    return response
def get_all_table_info(schema='public'):
    all_table_infos = {each:get_column_names(each) for each in get_all_table_names()}
    return all_table_infos
def get_value_from_row(row):
    if isinstance(row,list):
        for i,item in enumerate(row):
            if isinstance(item,dict):
                item = list(item.values())
            row[i] = item[0]
    if isinstance(row,dict):
        row = list(row.values())
        row = row[0]
    return row[0]
def fetch_combo_result(*arg,**kwargs):
    response = fetch_any_combo(*arg,**kwargs)
    result = get_value_from_row(response)
    return result



from abstract_utilities import make_list,SingletonMeta,get_logFile
import psycopg2
from psycopg2.extras import RealDictCursor
from abstract_flask import initialize_call_log
from abstract_utilities import make_list,SingletonMeta
from psycopg2 import sql, connect
from abstract_security import get_env_value
import traceback
import warnings
from .....managers.connectionManager.utils import connect_db
logger = get_logFile('fetch_utils')



from ..imports import *
# Environment setup
ENV_PATH = '/home/solcatcher/.env'

def get_env_val(key, default=None):
    """Retrieve environment variable with optional default."""
    return get_env_value(key=key, path=ENV_PATH) or default
def manual_connecty():
    # Centralized configuration
    CONFIG = {
        'RABBITMQ': {
            'HOST': get_env_val("SOLCATCHER_AMQP_HOST", 'local'),
            'PORT': get_env_val("SOLCATCHER_AMQP_PORT", '5672'),
            'USER': get_env_val("SOLCATCHER_AMQP_USER", 'solcatcher'),
            'NAME': get_env_val("SOLCATCHER_AMQP_NAME", 'solcatcher'),
            'PASSWORD': get_env_val("SOLCATCHER_AMQP_PASSWORD", 'solcatcher123'),
        },
        'POSTGRESQL': {
            'HOST': get_env_val("SOLCATCHER_POSTGRESQL_HOST", 'solcatcher.io'),
            'PORT': get_env_val("SOLCATCHER_POSTGRESQL_PORT", '5432'),
            'USER': get_env_val("SOLCATCHER_POSTGRESQL_USER", 'solcatcher'),
            'NAME': get_env_val("SOLCATCHER_POSTGRESQL_NAME", 'solcatcher'),
            'PASSWORD': get_env_val("SOLCATCHER_POSTGRESQL_PASSWORD", 'solcatcher123!!!456'),
        }
    }

    DB_URL = (
        f"postgresql://{CONFIG['POSTGRESQL']['USER']}:{CONFIG['POSTGRESQL']['PASSWORD']}"
        f"@{CONFIG['POSTGRESQL']['HOST']}:{CONFIG['POSTGRESQL']['PORT']}/{CONFIG['POSTGRESQL']['NAME']}"
    )

    def get_connection():
        """Establish a PostgreSQL connection."""
        return psycopg2.connect(
            dbname=CONFIG['POSTGRESQL']['NAME'],
            user=CONFIG['POSTGRESQL']['USER'],
            password=CONFIG['POSTGRESQL']['PASSWORD'],
            host=CONFIG['POSTGRESQL']['HOST'],
            port=CONFIG['POSTGRESQL']['PORT']
        )


import logging
import psycopg2
from psycopg2.extras import RealDictCursor
from psycopg2 import sql
from abstract_security import get_env_value
import traceback
import warnings
from abstract_utilities import get_logFile
from ...managers.columnNamesManager.utils.main import columnNamesManager,query_data,get_all_table_names
from ...managers.connectionManager.utils import connectionManager,get_cur_conn

